__all__ = [ "project" ]

project = {
    "name"         : "DoubleRatchet",
    "description"  : "A Python implementation of the Double Ratchet algorithm.",
    "url"          : "https://github.com/Syndace/python-doubleratchet",
    "year"         : "2024",
    "author"       : "Tim Henkes (Syndace)",
    "author_email" : "me@syndace.dev",
    "categories"   : [
        "Topic :: Security :: Cryptography"
    ]
}
